@@ -57,6 +57,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
                  endog,
                  exog,
                  classifier_pred, 
+                 classifier_true_ind=None,
                  classifier_ind = None,
                  classifier_ind_name = None,
                  entity_effects = False,
@@ -67,12 +68,16 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
                  check_rank=True,
                  singletons = True,
                  check_absorbed=True,
+                 cm=None,
                  **kwargs):
         
         print("Initializing...")    
         
-        
-        self.check_absorbed= check_absorbed
+        self.cm = cm
+        self.check_absorbed= check_absorbed 
+        self.classifier_true_ind = classifier_true_ind
+        if classifier_true_ind is not None:
+            self._classifier_true_ind_name = classifier_true_ind.name
 
         if classifier_ind is None:
             self._group = endog.index.get_level_values(2).name
@@ -88,14 +93,15 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             raise Exception("If either `classifier_ind` or `classifier_ind_name` is defined, the other must be defined as well, but ")
                 
         self.classifier_pred = classifier_pred
-        self.classifier_map = {k:v for k,v in zip(self.classifier_pred.columns, self.classifier_ind.cat.categories)}
+        if self.classifier_pred is not None:
+            self.classifier_map = {k:v for k,v in zip(self.classifier_pred.columns, self.classifier_ind.cat.categories)}
         
         endog_cat = (
             endog
             .assign(gaul_class = self.ind)
             .assign(gaul = self.classifier_ind)
-            .query(f"{self._group}==gaul_class")
-            .drop(columns=[self._group, 'gaul_class'])
+            .query(f"gaul==gaul_class")
+            .drop(columns=['gaul', 'gaul_class'])
             )
         
         
@@ -103,8 +109,8 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             exog
             .assign(gaul_class = self.ind)
             .assign(gaul = self.classifier_ind)
-            .query(f"{self._group}==gaul_class")
-            .drop(columns=[self._group, 'gaul_class'])
+            .query(f"gaul==gaul_class")
+            .drop(columns=['gaul', 'gaul_class'])
             )
 
         self.p_model = PanelOLS(dependent=endog_cat,
@@ -173,13 +179,15 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         cls,
         formula,
         data,
-        classifier_pred, 
+        classifier_pred=None, 
+        classifier_true_ind=None,
         other_effects=None,
         weights=None,
         drop_absorbed=False,
         check_rank=True,
         singletons = True,
         check_absorbed=True,
+        cm=None,
         **kwargs
     ):
         
@@ -188,6 +196,14 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         if classifier_ind_name == 'gaul':
             classifier_ind = data[classifier_ind_name].unique().tolist()
             classifier_ind = _reorganize_gaul_ind(classifier_ind)
+        else:
+            classifier_ind = data[classifier_ind_name].unique().tolist()
+            
+        if classifier_pred is None:
+            print("individual level predictions not provided, will only use confusion matrix")
+            
+        if classifier_pred is None and classifier_true_ind is None:
+            raise Exception("Must specify either classifier_pred or classifier_true_ind")
         
         parser = PanelFormulaParser(formula, data)
         
@@ -197,6 +213,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             endog = endog,
             exog = exog,
             classifier_pred=classifier_pred, 
+            classifier_true_ind = classifier_true_ind,
             classifier_ind = data[classifier_ind_name].astype(pd.CategoricalDtype(classifier_ind, ordered=True)),
             classifier_ind_name = classifier_ind_name,
             entity_effects = parser.entity_effect,
@@ -207,6 +224,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             check_rank=check_rank,
             singletons = singletons,
             check_absorbed=check_absorbed,
+            cm=cm,
             **kwargs
         )
                 
@@ -273,11 +291,17 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
     
     @functools.cached_property
     def ind(self):
-        return self.classifier_pred.idxmax(axis=1).map(self.classifier_map)
+        if self.classifier_pred is not None:
+            return self.classifier_pred.idxmax(axis=1).map(self.classifier_map)
+        else:
+            return self.classifier_true_ind.reset_index(self._group, drop=True)
     
     @functools.cached_property
     def p(self):
-        return self.classifier_pred.values
+        if self.classifier_pred is not None:
+            return self.classifier_pred.values
+        else:
+            return np.ones((self.y.shape[0], self.n_regimes))
     
     @functools.cached_property
     def p_slow(self):
@@ -307,146 +331,152 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
 
     def _ll(self, params):
                 
-        
-        if self.n_regimes == 4:
-            cm = np.array(
-                [
-                    [0.85400498, 0.17394031, 0.0470923, 0.17593651],
-                    [0.07902572, 0.71988785, 0.13576904, 0.04206538],
-                    [0.00444184, 0.05345483, 0.7683022, 0.01397825],
-                    [0.06252746, 0.05271701, 0.04883646, 0.76801986],
-                ]
-            )
+        if self.cm is None:
+            if self.n_regimes == 4:
+                cm = np.array(
+                    [
+                        [0.85400498, 0.17394031, 0.0470923, 0.17593651],
+                        [0.07902572, 0.71988785, 0.13576904, 0.04206538],
+                        [0.00444184, 0.05345483, 0.7683022, 0.01397825],
+                        [0.06252746, 0.05271701, 0.04883646, 0.76801986],
+                    ]
+                )
 
-        elif self.n_regimes == 10:
-            cm = np.array(
-                [
+            elif self.n_regimes == 10:
+                cm = np.array(
                     [
-                        7.51733523e-01,
-                        3.36037080e-02,
-                        1.06639147e-02,
-                        1.33630290e-02,
-                        9.74529347e-02,
-                        7.45621493e-02,
-                        5.60966108e-03,
-                        2.33992767e-03,
-                        2.18987342e-02,
-                        5.46599963e-03,
-                    ],
-                    [
-                        1.19240490e-02,
-                        8.86442642e-01,
-                        1.71998624e-03,
-                        5.93912398e-04,
-                        1.60149928e-02,
-                        1.70039109e-03,
-                        3.11647838e-04,
-                        0.00000000e00,
-                        2.53164557e-03,
-                        2.77932185e-04,
-                    ],
-                    [
-                        1.15352213e-02,
-                        4.05561993e-03,
-                        5.24251806e-01,
-                        2.87552586e-02,
-                        2.17224636e-02,
-                        1.30079918e-02,
-                        4.37086093e-02,
-                        7.53031270e-02,
-                        8.17721519e-02,
-                        1.17750602e-01,
-                    ],
-                    [
-                        8.73566198e-02,
-                        1.62224797e-02,
-                        1.24183007e-01,
-                        7.07745608e-01,
-                        3.45003833e-02,
-                        1.86958000e-01,
-                        1.97506817e-01,
-                        1.21676239e-01,
-                        6.70886076e-02,
-                        1.13118399e-01,
-                    ],
-                    [
-                        2.98749271e-02,
-                        3.36037080e-02,
-                        5.84795322e-03,
-                        1.68275179e-03,
-                        7.81412386e-01,
-                        3.65584084e-03,
-                        1.63615115e-03,
-                        1.27632419e-03,
-                        2.83544304e-02,
-                        2.03816935e-03,
-                    ],
-                    [
-                        8.38571706e-02,
-                        1.50637312e-02,
-                        2.20158239e-02,
-                        4.91957436e-02,
-                        1.95928103e-02,
-                        6.75990478e-01,
-                        1.19984418e-02,
-                        3.40353116e-03,
-                        1.75949367e-02,
-                        1.29701686e-02,
-                    ],
-                    [
-                        5.05475990e-03,
-                        4.05561993e-03,
-                        5.02235982e-02,
-                        1.68225687e-01,
-                        1.44816424e-03,
-                        2.27852406e-02,
-                        4.15894040e-01,
-                        1.66985748e-01,
-                        3.79746835e-03,
-                        1.37669075e-01,
-                    ],
-                    [
-                        1.68491997e-03,
-                        0.00000000e00,
-                        1.61678707e-02,
-                        7.72086117e-03,
-                        1.70372263e-04,
-                        3.48580173e-03,
-                        1.18426178e-01,
-                        4.69049138e-01,
-                        1.64556962e-03,
-                        6.95756902e-02,
-                    ],
-                    [
-                        1.09519798e-02,
-                        5.79374276e-03,
-                        2.23598211e-02,
-                        3.46448899e-03,
-                        2.53002811e-02,
-                        3.74086040e-03,
-                        1.16867939e-03,
-                        2.12720698e-04,
-                        7.68481013e-01,
-                        4.53955901e-03,
-                    ],
-                    [
-                        6.02682911e-03,
-                        1.15874855e-03,
-                        2.22566219e-01,
-                        1.92526602e-02,
-                        2.38521169e-03,
-                        1.41132460e-02,
-                        2.03739774e-01,
-                        1.59753244e-01,
-                        6.83544304e-03,
-                        5.36594404e-01,
-                    ],
-                ]
-            )
+                        [
+                            7.51733523e-01,
+                            3.36037080e-02,
+                            1.06639147e-02,
+                            1.33630290e-02,
+                            9.74529347e-02,
+                            7.45621493e-02,
+                            5.60966108e-03,
+                            2.33992767e-03,
+                            2.18987342e-02,
+                            5.46599963e-03,
+                        ],
+                        [
+                            1.19240490e-02,
+                            8.86442642e-01,
+                            1.71998624e-03,
+                            5.93912398e-04,
+                            1.60149928e-02,
+                            1.70039109e-03,
+                            3.11647838e-04,
+                            0.00000000e00,
+                            2.53164557e-03,
+                            2.77932185e-04,
+                        ],
+                        [
+                            1.15352213e-02,
+                            4.05561993e-03,
+                            5.24251806e-01,
+                            2.87552586e-02,
+                            2.17224636e-02,
+                            1.30079918e-02,
+                            4.37086093e-02,
+                            7.53031270e-02,
+                            8.17721519e-02,
+                            1.17750602e-01,
+                        ],
+                        [
+                            8.73566198e-02,
+                            1.62224797e-02,
+                            1.24183007e-01,
+                            7.07745608e-01,
+                            3.45003833e-02,
+                            1.86958000e-01,
+                            1.97506817e-01,
+                            1.21676239e-01,
+                            6.70886076e-02,
+                            1.13118399e-01,
+                        ],
+                        [
+                            2.98749271e-02,
+                            3.36037080e-02,
+                            5.84795322e-03,
+                            1.68275179e-03,
+                            7.81412386e-01,
+                            3.65584084e-03,
+                            1.63615115e-03,
+                            1.27632419e-03,
+                            2.83544304e-02,
+                            2.03816935e-03,
+                        ],
+                        [
+                            8.38571706e-02,
+                            1.50637312e-02,
+                            2.20158239e-02,
+                            4.91957436e-02,
+                            1.95928103e-02,
+                            6.75990478e-01,
+                            1.19984418e-02,
+                            3.40353116e-03,
+                            1.75949367e-02,
+                            1.29701686e-02,
+                        ],
+                        [
+                            5.05475990e-03,
+                            4.05561993e-03,
+                            5.02235982e-02,
+                            1.68225687e-01,
+                            1.44816424e-03,
+                            2.27852406e-02,
+                            4.15894040e-01,
+                            1.66985748e-01,
+                            3.79746835e-03,
+                            1.37669075e-01,
+                        ],
+                        [
+                            1.68491997e-03,
+                            0.00000000e00,
+                            1.61678707e-02,
+                            7.72086117e-03,
+                            1.70372263e-04,
+                            3.48580173e-03,
+                            1.18426178e-01,
+                            4.69049138e-01,
+                            1.64556962e-03,
+                            6.95756902e-02,
+                        ],
+                        [
+                            1.09519798e-02,
+                            5.79374276e-03,
+                            2.23598211e-02,
+                            3.46448899e-03,
+                            2.53002811e-02,
+                            3.74086040e-03,
+                            1.16867939e-03,
+                            2.12720698e-04,
+                            7.68481013e-01,
+                            4.53955901e-03,
+                        ],
+                        [
+                            6.02682911e-03,
+                            1.15874855e-03,
+                            2.22566219e-01,
+                            1.92526602e-02,
+                            2.38521169e-03,
+                            1.41132460e-02,
+                            2.03739774e-01,
+                            1.59753244e-01,
+                            6.83544304e-03,
+                            5.36594404e-01,
+                        ],
+                    ]
+                )
+        else:
+            cm = self.cm.astype(float)
+
         
         # generate X so that we can for absorbed variables
         X = self.X
         
+        if X.ndim ==1:
+            raise Exception("Regimes not of equal size...")
+        
         results_df = pd.Series(params, index=self._param_order)
         
         beta_df = results_df.loc[results_df.index != 'sigma']
@@ -469,10 +499,8 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         # get row_maxes for probabilities
         row_maxes = self.p.max(axis=1, keepdims=True)
         class_ind = np.where(self.p==row_maxes, self.p, 0)
-    
         return np.log((rnl*(class_ind@cm.T)).sum(axis=1))
 
-
     def _params_to_ll(self, params):
 
         # Since we always put the covariates at the end, take them out from the end
